from abc import ABC
from dataclasses import dataclass

from typing_extensions import Self


class HardwareException(IOError):
	pass

class OpeningPortException(HardwareException):
	pass

class SendingException(HardwareException):
	pass

class ReadingException(HardwareException):
	pass

class TimeoutException(ReadingException):
	pass

@dataclass
class RawFrame:
	identifier: int
	data: bytes

@dataclass
class HardwarePort:
	port: str
	port_name: str # for /dev/ttyUSB0, this would be ttyUSB0
	serial_number: str = None
	manufacturer: str = None

	def description (self) -> str:
		return '{} - {}'.format(
			self.port,
			' '.join([x for x in [self.manufacturer, self.serial_number] if x is not None])
		)

class HardwareABC(ABC):
	'''
	Abstract hardware class
	'''
	def __init__ (self, port: str, *args) -> None:
		'''
		Create a new hardware instance, assign arguments. No logic is executed here

		:param port: string identifier of the port, for example /dev/ttyUSB0 or can0
		:return:
		'''
		self.port: str = port
		self._port_opened: bool = False

	def open (self) -> bool:
		'''
		Open the port

		:return: a boolean indicating whether the port was opened successfully
		:rtype: bool
		'''
		pass

	def is_open (self) -> bool:
		'''
		Check whether the hardware socket is open

		:return: a boolean indicating whether the hardware socket/port is opened 
		'''
		return self._port_opened

	def read (self, length: int) -> RawFrame:
		'''
		Read X bytes from the buffer

		:param length: how many bytes to read
		:return: read frame. The data will never be smaller than provided length because in such case a TimeoutException would be thrown
		:rtype: RawFrame
		'''
		pass

	def write (self, data: RawFrame) -> int:
		'''
		Write to the port
		
		:param data: bytes to write
		:return: number of bytes written
		:rtype: int
		'''
		pass

	def close (self) -> None:
		'''
		Close the port
		'''
		pass

	def set_baudrate (self, baudrate: int) -> Self:
		'''
		Set hardware port baudrate - where applicable

		:param int baudrate: requested baudrate in bits per second
		:rtype: :class:`Self`
		'''
		pass

	def get_baudrate (self) -> int:
		'''
		Get current hardware port baudrate

		:return: baudrate in bits per second
		:rtype: int
		'''
		return self.baudrate

	def set_timeout (self, timeout: float) -> Self:
		'''
		Set maximum timeout for reading

		:param float timeout: Timeout in seconds
		:rtype: Self
		'''
		pass

	def get_timeout (self) -> float:
		'''
		Get the maximum timeout for reading

		:return: timeout in seconds
		:rtype: float
		'''
		return self.timeout

	@staticmethod
	def available_ports () -> list[HardwarePort]:
		'''
		Enumerate available ports

		:return: list of :py:class:`HardwarePort` objects
		:rtype: list[HardwarePort]
		'''
		pass

	def __del__ (self) -> None:
		self.close()


from dataclasses import dataclass
from enum import Enum, auto


class CcpReturnCodeCategory (Enum):
	'''
	Return codes in CCP are grouped into categories. 
	SUCCESS category doesn't exist in theory, 0x00 ACKNOWLEDGE 
	response doesn't have a category assigned. SUCCESS was added 
	for code clarity

	'''
	SUCCESS = auto()
	'''Added for code clarity'''

	WARNING = auto()
	'''C0 - Warning'''

	SPURIOUS = auto()
	'''C1 - Spurious (comm error, busy, ...). Action: wait (ACK or timeout)'''

	RESOLVABLE = auto()
	'''C2 - Resolvable (temp. power loss, ...). Action: reinitialize'''

	UNRESOLVABLE = auto()
	'''C3 - Unresolvable (setup, overload, ...). Action: terminate'''

class CcpReturnCodeCategoryMap (Enum):
	ACKNOWLEDGE = CcpReturnCodeCategory.SUCCESS
	DAQ_PROCESSOR_OVERLOAD = CcpReturnCodeCategory.WARNING
	COMMAND_PROCESSOR_BUSY = CcpReturnCodeCategory.SPURIOUS
	DAQ_PROCESSOR_BUSY = CcpReturnCodeCategory.SPURIOUS
	INTERNAL_TIMEOUT = CcpReturnCodeCategory.SPURIOUS
	KEY_REQUEST = CcpReturnCodeCategory.SPURIOUS
	SESSION_STATUS_REQUEST = CcpReturnCodeCategory.SPURIOUS
	COLD_START_REQUEST = CcpReturnCodeCategory.RESOLVABLE
	CAL_DATA_INIT_REQUEST = CcpReturnCodeCategory.RESOLVABLE
	DAQ_LIST_INIT_REQUEST = CcpReturnCodeCategory.RESOLVABLE
	CODE_UPDATE_REQUEST = CcpReturnCodeCategory.RESOLVABLE
	UNKNOWN_COMMAND = CcpReturnCodeCategory.UNRESOLVABLE
	COMMAND_SYNTAX = CcpReturnCodeCategory.UNRESOLVABLE
	PARAMETERS_OUT_OF_RANGE = CcpReturnCodeCategory.UNRESOLVABLE
	ACCESS_DENIED = CcpReturnCodeCategory.UNRESOLVABLE
	OVERLOAD = CcpReturnCodeCategory.UNRESOLVABLE
	ACCESS_LOCKED = CcpReturnCodeCategory.UNRESOLVABLE
	RESOURCE_NOT_AVAILABLE = CcpReturnCodeCategory.UNRESOLVABLE

class CcpReturnCode (Enum):
	ACKNOWLEDGE = 0x00
	DAQ_PROCESSOR_OVERLOAD = 0x01
	COMMAND_PROCESSOR_BUSY = 0x10
	DAQ_PROCESSOR_BUSY = 0x11
	INTERNAL_TIMEOUT = 0x12
	KEY_REQUEST = 0x18
	SESSION_STATUS_REQUEST = 0x19
	COLD_START_REQUEST = 0x20
	CAL_DATA_INIT_REQUEST = 0x21
	DAQ_LIST_INIT_REQUEST = 0x22
	CODE_UPDATE_REQUEST = 0x23
	UNKNOWN_COMMAND = 0x30
	COMMAND_SYNTAX = 0x31
	PARAMETERS_OUT_OF_RANGE = 0x32
	ACCESS_DENIED = 0x33
	OVERLOAD = 0x34
	ACCESS_LOCKED = 0x35
	RESOURCE_NOT_AVAILABLE = 0x36

	@property
	def category(self) -> CcpReturnCodeCategory:
		return CcpReturnCodeCategoryMap[self.name].value

	@property
	def success (self) -> bool:
		return self.category == CcpReturnCodeCategory.SUCCESS

	def __str__ (self) -> str:
		return 'CcpReturnCode(code={}, name={}, category={})'.format(hex(self.value), self.name, self.category.name)

	def __repr__ (self) -> str:
		return self.__str__()

@dataclass
class CcpResponseFrame:
	'''
	Also known as Data Transmission Object - DTO. 
	Perhaps this should be renamed
	
	Structure:
	1 byte - data packet ID = PID.
		0xFF - DTO contains a Command Return Message
		0xFE - DTO contains an Event Message. Counter is ignored
		0 < x < 0xFD - DTO contains a Data Acquisition Message corresponding to ODT n
	1 byte - status - command return / error code
	1 byte - counter of transmitted frames (from slave to master)
	5 bytes - data 
	'''
	packet_id: int
	status: int
	counter: int
	data: bytes

	def __str__ (self) -> str:
		return 'CcpResponseFrame({})'.format(' '.join([hex(x)[2:].zfill(2) for x in list(self.data)]))

	def __repr__ (self) -> str:
		return self.__str__()

@dataclass
class CcpResponse:
	return_code: CcpReturnCode
	frame: CcpResponseFrame 

	def success (self) -> bool: # @todo: should this be removed, considering we have it already in return_code?
		return self.return_code.success

	def get_data (self) -> bytes: # todo: deprecate?
		return self.frame.data



import struct

from typing_extensions import Self

from .enums import *
from .kwp2000_command import Kwp2000Command, Kwp2000CommandWithSubservices


class AccessTimingParameters(Kwp2000CommandWithSubservices): # TimingParameterIdentifier
	service_identifier = 0x83

	def read_limits_of_possible_timing_parameters (self) -> Self:
		return self.set_subservice_identifier(TimingParameterIdentifier.READ_LIMITS_OF_POSSIBLE_TIMING_PARAMETERS.value)

	def set_timing_parameters_to_default_values (self) -> Self:
		return self.set_subservice_identifier(TimingParameterIdentifier.SET_TIMING_PARAMETERS_TO_DEFAULT_VALUES.value)

	def read_currently_active_timing_parameters (self) -> Self:
		return self.set_subservice_identifier(TimingParameterIdentifier.SET_TIMING_PARAMETERS_TO_GIVEN_VALUES.value)

	def set_timing_parameters_to_given_values (self,
			p2min: int,
			p2max: int,
			p3min: int,
			p3max: int,
			p4min: int
		) -> Self:
		self.set_subservice_identifier(TimingParameterIdentifier.SET_TIMING_PARAMETERS_TO_GIVEN_VALUES.value)
		return self.append_data(bytes([p2min, p2max, p3min, p3max, p4min]))

class ClearDiagnosticInformation(Kwp2000Command):
	service_identifier = 0x14

class ControlDTCSetting(Kwp2000Command):
	service_identifier = 0x85

class DisableNormalMessageTransmission(Kwp2000Command):
	service_identifier = 0x28

	def init(self, response_type: ResponseType) -> None:
		self.set_data(bytes([response_type.value]))

class DynamicallyDefineLocalIdentifier(Kwp2000Command):
	service_identifier = 0x2C

class ECUReset(Kwp2000Command):
	'''
	This service requests the ECU to effectively perform a reset 
	based on the content of the reset_mode parameter. 
	The reset_mode parameter may specify a reset for the entire ECU 
	or selective portions of on-board memory
	'''
	service_identifier = 0x11

	def init (self, reset_mode: ResetMode) -> None:
		self.set_data(bytes([reset_mode.value]))

class EnableNormalMessageTransmission(Kwp2000Command):
	service_identifier = 0x29

	def init (self, response_type: ResponseType) -> None:
		self.set_data(bytes([response_type.value]))

class InputOutputControlByLocalIdentifier(Kwp2000Command):
	service_identifier = 0x30

	def init (self, control_identifier: int, control_parameter: InputOutputControlParameter, *control_state) -> None:
		self.set_data(bytes([control_identifier, control_parameter.value, *control_state]))

class ReadDataByIdentifier(Kwp2000Command):
	service_identifier = 0x22

class ReadDataByLocalIdentifier(Kwp2000Command):
	service_identifier = 0x21

	def init (self, record_local_identifier: int) -> None:
		self.set_data(bytes([record_local_identifier]))

class ReadDTCsByStatus(Kwp2000Command):
	service_identifier = 0x18

	def init (self, status: DtcStatus, group: DtcGroup) -> None:
		group = struct.pack('>H', group.value)
		self.set_data(bytes([status.value, *group]))

class ReadEcuIdentification(Kwp2000Command):
	service_identifier = 0x1A

	def init (self, identifier) -> None:
		self.set_data(bytes([identifier]))

class ReadMemoryByAddress(Kwp2000Command):
	service_identifier = 0x23

	def init (self, offset: int = 0x000000, size: int = 0xFE) -> None:
		address = struct.pack('>L', offset)[1:]

		self.set_data(bytes([*address, size]))

class ReadStatusOfDTC(Kwp2000Command):
	service_identifier = 0x01

	def init (self, dtc) -> None:
		self.set_data(bytes([dtc]))

class RequestDownload(Kwp2000Command):
	service_identifier = 0x34

	def init (self, 
			offset: int, 
			compression_type: CompressionType,
			encryption_type: EncryptionType,
			size: int
		) -> None:
		address = struct.pack('>L', offset)[1:]
		data_format = (compression_type.value << 4) | encryption_type.value
		size = struct.pack('>L', size)[1:]

		self.set_data(bytes([*address, data_format, *size]))

class RequestRoutineResultsByLocalIdentifier(Kwp2000Command):
	service_identifier = 0x33

	def init (self, routine_identifier: int) -> None:
		self.set_data(bytes([routine_identifier]))

class RequestTransferExit(Kwp2000Command):
	service_identifier = 0x37

class RequestUpload(Kwp2000Command):
	service_identifier = 0x35

	def init (self, 
			offset: int, 
			compression_type: CompressionType,
			encryption_type: EncryptionType,
			size: int
		) -> None:
		address = struct.pack('>L', offset)[1:]
		data_format = (compression_type.value << 4) | encryption_type.value
		size = struct.pack('>L', size)[1:]

		self.set_data(bytes([*address, data_format, *size]))

class ResponseOnEvent(Kwp2000Command):
	service_identifier = 0x86

class SecurityAccess(Kwp2000CommandWithSubservices):
	service_identifier = 0x27

	def request_seed (self, access_level: int = AccessType.PROGRAMMING_REQUEST_SEED.value) -> Self:
		return self.set_subservice_identifier(access_level)

	def send_key (self, key: bytes, access_level: int = AccessType.PROGRAMMING_SEND_KEY.value) -> Self:
		self.set_subservice_identifier(access_level)
		return self.append_data(key)

class StartCommunication(Kwp2000Command):
	service_identifier = 0x81

class StartRoutineByLocalIdentifier(Kwp2000Command):
	service_identifier = 0x31

	def init (self, routine_identifier: int, *routine_entry_option) -> None:
		self.set_data(bytes([routine_identifier, *routine_entry_option]))

class StartDiagnosticSession(Kwp2000Command):
	'''
	Start diagnostic session on the ECU. 

	- By default, Normal/Default session (0x81) is automatically enabled by the ECU,
	but not if the ECU entered the Passive Session (0x90) before powering down.
	- ECU will return to Normal/Default session (0x81) after timeout 
	of another diagnostic session 
	- A positive response will be returned if the requested session was already active
	- ECU will stay in it's previous session if the requested session is 
	unable to be activated
	- Sessions Flash Reprogramming (0x85), Stand By (0x89), Extended Diagnostic (0x92)
	must be kept active with TesterPresent (0x3E) service if no other diagnostic 
	messages are sent within P3 max (see timing parameters)
	'''

	service_identifier = 0x10

	def init (self, session_type: DiagnosticSession, dev_baudrate_identifier: int | None = None) -> None:
		self.set_data(bytes([session_type.value]))
		# while it's not present in any public documentations, most ECU manufacturers use 
		# the second parameter of StartDiagnosticSession as a baudrate switch. 
		# for example, 0x03 on SIMK43 will result in 40k baud, 0x04 - 60k 
		if (dev_baudrate_identifier): 
			self.append_data(bytes([dev_baudrate_identifier]))

class StopCommunication(Kwp2000Command):
	service_identifier = 0x82

class StopRoutineByLocalIdentifier(Kwp2000Command):
	service_identifier = 0x32

	def init (self, routine_identifier: int, *routine_exit_option) -> None:
		self.set_data(bytes([routine_identifier, *routine_exit_option]))

class TesterPresent(Kwp2000Command):
	service_identifier = 0x3E

	def init (self, response_type: ResponseType) -> None:
		self.set_data(bytes([response_type.value]))

class TransferData(Kwp2000Command):
	service_identifier = 0x36

class WriteDataByIdentifier(Kwp2000Command):
	service_identifier = 0x2E

class WriteDataByLocalIdentifier(Kwp2000Command):
	service_identifier = 0x3B

	def init (self, record_local_identifier: int, record_value: bytes) -> None:
		self.set_data(bytes([record_local_identifier]) + record_value)

class WriteMemoryByAddress(Kwp2000Command):
	service_identifier = 0x3D

	def init (self, offset: int, data_to_write: bytes) -> None:
		size = len(data_to_write)

		address = struct.pack('>L', offset)[1:]

		self.set_data(bytes([*address, size]) + data_to_write)

class StopDiagnosticSession(Kwp2000Command):
	service_identifier = 0x20


import logging
import time

import serial
import serial.tools.list_ports
from typing_extensions import Self

from .hardware_abc import (
	HardwareABC,
	HardwarePort,
	OpeningPortException,
	RawFrame,
	TimeoutException,
)

from ..utils import ns_to_ms, ms_to_ns

logger = logging.getLogger(__name__)

class KLineHardware(HardwareABC):
	'''
	Hardware class for serial devices, using pyserial as a backend
	'''

	def __init__ (self, port: str, baudrate: int = 10400, timeout: float = 2) -> None:
		self.port, self.baudrate = port, baudrate
		self.timeout = timeout
		self._port_opened = False
		self.socket: serial.Serial = None

	def open (self) -> bool:
		try:
			self.socket = serial.Serial(self.port, self.baudrate, timeout=self.timeout)
			self._port_opened = True
		except serial.serialutil.SerialException as e:
			raise OpeningPortException(e)
		
		self._reset_adapter()
		self._set_kline_mode()

		return True

	def read (self, length: int) -> RawFrame:
		message = self.socket.read(length)

		if (len(message) < length):
			raise TimeoutException

		return RawFrame(identifier=False, data=message)

	def write (self, frame: RawFrame) -> int:
		# @todo - this timeout is needed for now, otherwise 
		# we'll get a timeout when trying to start diagnostic session
		# find out _why_ is it needed - some magic value instead 
		# of something calculated shouldnt be required here
		time.sleep((50/1000)*2)
		data = frame.data
		bytes_written = self.socket.write(data)

		while self.socket.out_waiting > 0:
			time.sleep(0.001)

		echo = self.socket.read(bytes_written)
		if (echo != data):
			logger.error('K-Line echo different than sent payload! \nPayload: {}\nEcho: {}'.format(
				' '.join([hex(x) for x in list(data)]),
				' '.join([hex(x) for x in list(echo)])
			))

		return bytes_written

	def close (self) -> None:
		if not self.is_open():
			logger.info('Tried to close an already closed port')
			return # should this be an exception?
		try:
			self.socket.break_condition = False
			self.socket.reset_input_buffer()
			self.socket.reset_output_buffer()
		except AttributeError as e:
			logger.info('Couldn\'t flush socket buffer before closing: {}'.format(str(e)))
		except serial.serialutil.PortNotOpenError:
			pass
			
		try:
			self.socket.close()
		except AttributeError as e:
			logger.error('Couldn\'t close socket: {}'.format(str(e)))
		except serial.serialutil.PortNotOpenError:
			pass

		self._port_opened = False

	def set_timeout (self, timeout: float) -> Self:
		self.socket.timeout = timeout
		self.timeout = timeout
		return self

	def set_baudrate (self, baudrate: int) -> Self:
		self.socket.baudrate = baudrate
		self.baudrate = baudrate
		self.socket.flush()
		return self

	def _reset_adapter (self) -> None:
		'''
		Flip the Data Terminal Ready state. 
		This is required on some of the knock-off adapters which will otherwise fail to perform Fast Init
		'''
		self.socket.setDTR(1)
		time.sleep(0.1)
		self.socket.setDTR(0)
		time.sleep(0.1)

	def _set_kline_mode (self) -> None:
		'''
		Set the adapter to KLine (KKL) mode by flipping DTR and RTS to 0.
		This is required on some of the knock-off adapters which will otherwise fail to perform Fast Init
		'''
		self.socket.setDTR(0)
		self.socket.setRTS(0)
		time.sleep(0.1)

	def iso14230_fast_init (self, payload: bytes, timing_offset_ms: int = 0) -> tuple[bytes, int, int]:
		'''
		Perform FastInit by bringing the bus down for 25ms and then up for 25ms followed by a payload

		:return: input buffer contents (up to 40 bytes), elapsed time in low position (ms), elapsed time in high position (ms)
		'''

		fastinit_time = ms_to_ns(25-timing_offset_ms)

		# FastInit low
		self.socket.break_condition = True
		start_time = time.perf_counter_ns()

		# this is commented for now, as it doesn't seem to provide anything
		# other than messing with the timing. if it indeed doesn't solve
		# any issue, it'll be removed completely
		#self.socket.flush()  # Ensure the break is sent immediately

		while (time.perf_counter_ns() - start_time) < fastinit_time:
			pass  # Busy-wait until 25 ms has passed
		elapsed_time_low = time.perf_counter_ns() - start_time

		# FastInit high
		self.socket.break_condition = False
		start_time = time.perf_counter_ns()

		# this is commented for now, as it doesn't seem to provide anything
		# other than messing with the timing. if it indeed doesn't solve
		# any issue, it'll be removed completely
		#self.socket.flush()  # Ensure the break is sent immediately

		while (time.perf_counter_ns() - start_time) < fastinit_time:
			pass  # Busy-wait until 25 ms has passed
		elapsed_time_high = time.perf_counter_ns() - start_time
		
		self.socket.write(bytes(payload))

		'''
		Sure-fire way to clear out the incoming buffer. 
		socket.reset_input_buffer() didn't provide reliable results, probably due to timing.
		Why are we clearing out the incoming buffer?
		On genuine FTDI chips and some of the knockoffs, there'd be 0x00 in the buffer,
		then the request echo, and StartCommunication response afterwards.
		...
		On some chips, there'd be no 0x00. On some chips, there'd be no echo.
		On some chips, the response would be completely malformed,
		on some chips it'd be missing first 1-4 bytes, which would offset
		the buffer and trigger "echo different than sent payload" errors.
		-- 
		After giving it some thought, we realized we don't care about the StartCommunication 
		response. We don't have to know if fastinit was successful, the upcoming
		StartDiagnosticSession request will tell us that. 

		Is this the perfect solution? Probably not, but we're dealing with non-perfect
		hardware, and we want to support all of it.
		'''
		response = self.socket.read(40)

		return response, ns_to_ms(elapsed_time_low), ns_to_ms(elapsed_time_high)

	@staticmethod
	def available_ports () -> list[HardwarePort]:
		devices = []
		for port in reversed(serial.tools.list_ports.comports()):
			devices.append(HardwarePort(port.device, port.name, port.manufacturer, port.serial_number))
		return devices


from typing_extensions import Self


class CcpCommand:
	code = 0x0

	def __init__ (self, *args: tuple[bytes], **kwargs: dict) -> None:
		self.parameters: dict = dict(enumerate(args)) | kwargs
		self.init(*args, **kwargs)

	def init (self, *args: tuple[bytes], **kwargs: dict) -> None:
		if len(args) == 0:
			self.data: bytes = bytes()
		else:
			self.data: bytes = args[0]
			
	def set_data (self, data: bytes = bytes()) -> Self:
		self.data = data
		return self

	def get_data (self) -> bytes:
		return self.data

	def set_code (self, code: int) -> Self:
		self.code = code
		return self

	def get_code (self) -> int:
		return self.code

	def get_parameters (self) -> dict:
		'''
		Get parameters the command was initialized with

		:return: A dictionary of parameters name-value
		'''
		return self.parameters

	def __str__ (self) -> str:
		return 'CCPCommand(code={}, data={}|{})'.format(
			hex(self.get_code()), 
			' '.join([hex(x) for x in list(self.get_data())]),
			', '.join(f'{k}={v}' for k, v in self.get_parameters().items())
			)

	def __repr__ (self) -> str:
		return self.__str__()


from dataclasses import dataclass


@dataclass
class Kwp2000ResponseFrame:
	'''
	Kwp2000 response frame. 
	First byte is status - if it's success, status is going to be request service identifier + 0x40
	If it's an error, status is going to be 0x7f, first byte of data is going to be service identifier,
	second byte of data will be error identifier
	'''
	status: int # for success, service identifier + 0x40. error = 0x7f
	data: bytes

	def __repr__ (self) -> str:
		return 'Kwp2000ResponseFrame(status={}, data={!r})'.format(hex(self.status), self.data)

@dataclass
class Kwp2000Response:
	frame: Kwp2000ResponseFrame 

	def success (self) -> bool:
		return self.frame.status != 0x7F

	def get_data (self) -> bytes: # todo: deprecate?
		return self.frame.data


'''
Protocols
'''

from .ccp import *
from .kwp2000 import *
from .protocol_abc import ProtocolABC

__all__ = ['ProtocolABC', 'ccp_command', 'ccp_protocol', 'kwp2000_command', 'kwp2000_negative_status', 'kwp2000_protocol', 'kwp2000_response']



'''
CAN Calibration Protocol
'''

from . import commands, enums, types
from .ccp_command import CcpCommand
from .ccp_protocol import CcpException, CcpNegativeResponseException, CcpProtocol

__all__ = ['CcpCommand', 'CcpException', 'CcpNegativeResponseException', 'CcpProtocol', 'commands', 'enums', 'types']


'''
KWP2000 protocol, also known as ISO 14230-3
'''

from . import commands, enums
from .kwp2000_command import Kwp2000Command
from .kwp2000_negative_status import Kwp2000NegativeStatusIdentifierEnum
from .kwp2000_protocol import Kwp2000Exception, Kwp2000NegativeResponseException, Kwp2000Protocol

__all__ = ['Kwp2000Command', 'Kwp2000Exception', 'Kwp2000NegativeResponseException', 'Kwp2000NegativeStatusIdentifierEnum', 'Kwp2000Protocol', 'commands', 'enums']



'''
Transport layers
'''

from .ccp_over_can_transport import CcpOverCanTransport
from .kwp2000_over_can_transport import Kwp2000OverCanTransport
from .kwp2000_over_kline_transport import Kwp2000OverKLineTransport
from .transport_abc import PacketDirection, RawPacket, TransportABC

__all__ = ['CcpOverCanTransport', 'Kwp2000OverCanTransport', 'Kwp2000OverKLineTransport', 'PacketDirection', 'RawPacket', 'TransportABC']



import logging
import threading
import time
from dataclasses import dataclass

from ...hardware import TimeoutException
from ...transport import Kwp2000OverKLineTransport
from ..protocol_abc import ProtocolABC, ProtocolException
from .kwp2000_command import Kwp2000Command
from .kwp2000_negative_status import Kwp2000NegativeStatus, Kwp2000NegativeStatusIdentifierEnum
from .kwp2000_response import Kwp2000Response, Kwp2000ResponseFrame

logger = logging.getLogger(__name__)

class Kwp2000Exception(ProtocolException):
	pass

class Kwp2000NegativeResponseException(Kwp2000Exception):
	def __init__ (self, status: Kwp2000NegativeStatus) -> None:
		self.status = status

	def __str__ (self) -> str:
		return '{} - {}'.format(hex(self.status.identifier), self.status.message)
	
	def __repr__ (self) -> str:
		return self.__str__()

@dataclass
class Kwp2000RequestFrame:
	service_identifier: int
	data: bytes

	def to_pdu (self) -> bytes:
		return self.service_identifier.to_bytes(1, 'big') + self.data

class Kwp2000Protocol (ProtocolABC):
	def open (self) -> bool:
		if not self.transport.hardware.is_open():
			return self.transport.hardware.open()

		return True

	def init (self, init_command: Kwp2000Command, keepalive_command: Kwp2000Command | None = None, keepalive_delay: float = 1.5) -> bool:
		if isinstance(self.transport, Kwp2000OverKLineTransport):
			self.open()
			
			init_payload = Kwp2000RequestFrame(init_command.get_service_identifier(), init_command.get_data())

			try:
				self.transport.hardware.set_timeout(0.4)
				for (response, time_low, time_high) in self.transport.init(init_payload.to_pdu()):
					logger.debug('K-Line FastInit: response: {!r}, time low: {}, time high: {}'.format(response, time_low, time_high))
					self.transport.read_pdu()
				self.transport.hardware.set_timeout(2)
			except TimeoutException:
				pass

			self.transport.hardware.set_timeout(2)
		else:
			self.transport.init()

		self._execute_lock = threading.Lock()

		if keepalive_command:
			self.keepalive_command = keepalive_command
			self.keepalive_delay = keepalive_delay
			self._keepalive_thread = None
			self._last_execution_time = time.time()
			self._keepalive_event = threading.Event()
			self._start_keepalive()

		return True # there should be some error checking - but at this stage its hard to determine whether we succeeded


	def execute (self, command: Kwp2000Command) -> Kwp2000Response:
		frame = Kwp2000RequestFrame(command.get_service_identifier(), command.get_data())

		with self._execute_lock:
			response_pdu = self.transport.send_read_pdu(data=frame.to_pdu())
			self._last_execution_time = time.time()

			response = self.handle_errors(
				Kwp2000Response(Kwp2000ResponseFrame(status=response_pdu[0], data=response_pdu[1:]))
			)

		return response

	def handle_errors (self, response: Kwp2000Response) -> Kwp2000Response:
		if response.success():
			return response

		if response.frame.data[1] == Kwp2000NegativeStatusIdentifierEnum.REQUEST_CORRECTLY_RECEIVED_RESPONSE_PENDING.value:
			logger.info('ECU is busy, request received, response pending.')

			response_pdu = self.transport.read_pdu()
			self._last_execution_time = time.time()
			return self.handle_errors(Kwp2000Response(Kwp2000ResponseFrame(status=response_pdu[0], data=response_pdu[1:])))

		raise Kwp2000NegativeResponseException(Kwp2000NegativeStatus(identifier=response.frame.data[1]))

	def _keepalive (self) -> None:
		'''
		Send the keepalive command if keepalive_delay seconds elapsed since the last transmitted frame
		'''
		try:
			while not self._keepalive_event.is_set():
				time.sleep(1)
				if not self.transport.hardware.is_open(): 
					# this is solving a problem introduced by time.sleep 
					# being at the beginning of the loop instead of the end
					# time.sleep at the beginning is solving another problem
					# i dont remember what problem.
					break
				elapsed_time = time.time() - self._last_execution_time if self._last_execution_time else float('inf')
				if elapsed_time >= self.keepalive_delay:
					try:
						self.execute(self.keepalive_command)
					except (TimeoutException, Kwp2000NegativeResponseException, AttributeError) as e:
						break
		except KeyboardInterrupt:
			pass
		del self._keepalive_thread

	def _start_keepalive (self) -> None:
		'''
		Start the keep-alive thread
		'''
		self._keepalive_event.clear()
		self._keepalive_thread = threading.Thread(target=self._keepalive)
		self._keepalive_thread.start()

	def _stop_keepalive (self) -> None:
		if hasattr(self, '_keepalive_event'):
			self._keepalive_event.set()
		if hasattr(self, '_keepalive_thread'):
			self._keepalive_thread.join()

	def close (self) -> None:
		self._stop_keepalive()
		if self.transport.hardware.is_open():
			self.transport.hardware.close()


from typing_extensions import Self


class Kwp2000Command:
	service_identifier: int = 0x0

	def __init__ (self, *args: tuple[bytes], **kwargs: dict) -> None:
		self.parameters: dict = dict(enumerate(args)) | kwargs
		self.init(*args, **kwargs)

	def init (self, *args: tuple[bytes], **kwargs: dict) -> None:
		if len(args) == 0:
			self.data = bytes()
		else:
			self.data = args[0]

	def set_data (self, data: bytes=bytes()) -> Self:
		self.data = data

		return self

	def get_data (self) -> bytes:
		return self.data

	def append_data (self, data: bytes=bytes()) -> Self:
		self.data += data
		return self

	def set_service_identifier (self, service_identifier: int) -> Self:
		self.service_identifier = service_identifier
		return self

	def get_service_identifier (self) -> int:
		return self.service_identifier

	def get_parameters (self) -> dict:
		'''
		Get parameters the command was initialized with

		:return: A dictionary of parameters name-value
		'''
		return self.parameters

	def __str__ (self) -> str:
		return '{}(sid={}, data={}|{})'.format(
			type(self).__name__,
			hex(self.get_service_identifier()), 
			' '.join([hex(x) for x in list(self.get_data())]),
			', '.join(f'{k}={hex(v)}' for k, v in self.get_parameters().items())
		)

	def __repr__ (self) -> str:
		return self.__str__()

class Kwp2000CommandWithSubservices(Kwp2000Command):
	def set_subservice_identifier (self, subservice_identifier: int) -> Self:
		'''
		Overwrite the first byte of data with new subservice identifier
		'''
		return self.set_data(bytes([subservice_identifier]) + self.get_data()[1:])

	def get_subservice_identifier (self) -> int:
		return self.get_data()[0]

	def get_data (self) -> bytes:
		'''
		Get command payload, with subservice identifier being the first byte
		'''

		return self.data

	def __str__ (self) -> str:
		return '{}.Subservice(sid={}, sub={}, data={}|{})'.format(
			type(self).__name__,
			hex(self.get_service_identifier()), 
			hex(self.get_subservice_identifier()),
			' '.join([hex(x) for x in list(self.get_data())]),
			', '.join(f'{k}={hex(v)}' for k, v in self.get_parameters().items())
		)



from abc import ABC
from dataclasses import dataclass
from enum import Enum

from typing_extensions import Self

from ..hardware.hardware_abc import HardwareABC


class PacketDirection(Enum):
	INCOMING = 0
	OUTGOING = 1

@dataclass
class RawPacket:
	'''
	Raw packet being sent over the hardware port

	:param direction: Direction of the packet - incoming/outgoing
	:type direction: PacketDirection
	:param data: Raw data being sent over the hardware port
	:type data: bytes
	:param timestamp: Timestamp in miliseconds
	:type timestamp: int
	'''

	direction: PacketDirection
	data: bytes
	timestamp: int

	def __str__ (self) -> str:
		direction = 'Incoming' if self.direction == PacketDirection.INCOMING else 'Outgoing'
		return 'RawPacket({}, ts={}, data={!r})'.format(direction, self.timestamp, self.data)

	def __repr__ (self) -> str:
		return self.__str__()

class TransportABC(ABC):
	'''
	Transport layer for protocols

	:param buffer: Buffer storing last X packets, X determined by 
		buffer_size
	:type buffer: list[RawPacket]
	:param buffer_size: Determines the size of the buffer.
		0 - unlimited buffer, None - no buffer
	:type buffer_size: int
	'''
	buffer: list[RawPacket] = [] # is this shared between all transports? @todo critical
	buffer_size: int = 20

	def __init__ (self, hardware: HardwareABC, tx_id: int | None = None, rx_id: int | None = None) -> None:
		'''
		TransportABC constructor

		:param hardware: HardwareABC instance
		:param tx_id: Default identifier to transmit on - optional, hardware specific
		:param rx_id: Default identifier to listen for - optional, hardware specific. 
			For CAN based transports, this will be set as filter in the interface
		'''
		self.hardware = hardware
		self.tx_id, self.rx_id = tx_id, rx_id
		self.buffer = []

	def init (self) -> bool:
		'''
		Get ready to transfer: initialize the underlaying hardware if not initialized already
		Calling this method is not required - it's just a helper to avoid manual hardware initialization
		'''
		if not self.hardware.is_open():
			return self.hardware.open()
		return True

	def send_pdu (self, data: bytes) -> int:
		'''
		Send a compiled PDU over hardware port
		'''
		pass

	def read_pdu (self) -> bytes:
		'''
		Read a PDU over hardware port
		'''
		pass

	def send_read_pdu (self, data: bytes) -> bytes:
		'''
		Send PDU and read the response
		'''
		self.send_pdu(data)
		return self.read_pdu()

	def set_tx_id (self, tx_id: int) -> Self:
		'''
		Set default identifier to transmit on - hardware specific

		:param tx_id: Identifier to transmit on
		'''
		self.hardware.set_tx_id(tx_id)
		self.tx_id = tx_id

	def get_tx_id (self) -> int:
		'''
		Get default identifier to transmit on
		'''
		return self.tx_id

	def get_rx_id (self, rx_id: int) -> int:
		'''
		Get current identifier to listen for. Depending on the hardware,
		a hardware filter should be created with this identifier,
		meaning no other messages will be received or processed
		'''
		return self.rx_id

	def set_buffer_size (self, buffer_size: int) -> Self:
		'''
		Set size of the buffer. 

		:param buffer_size: 0 for unlimited, None for no logging
		:type buffer_size: int
		:rtype: TransportABC
		'''
		self.buffer_size = buffer_size

	def get_buffer_size (self) -> int:
		return self.buffer_size

	def buffer_push (self, packet: RawPacket) -> Self:
		if (self.buffer_size == None):
			return self

		self.buffer.append(packet)
		if (self.buffer_size != 0 and len(self.buffer) > self.buffer_size):
			del self.buffer[0:(len(self.buffer)-self.buffer_size)]
		
		return self

	def buffer_dump (self) -> list[RawPacket]:
		'''
		Retrieve contents of the buffer and empty the buffer

		:return: A list of RawPacket objects
		'''
		packets = self.buffer
		self.buffer = []
		return packets


'''
Hardware layers - concrete implementations of devices
'''

from .can_hardware import CanFilter, CanHardware
from .hardware_abc import (
    HardwareABC,
    HardwareException,
    HardwarePort,
    OpeningPortException,
    RawFrame,
    ReadingException,
    SendingException,
    TimeoutException,
)
from .kline_hardware import KLineHardware

__all__ = ['CanFilter', 'CanHardware', 'HardwareABC', 'HardwareException', 'HardwarePort', 'KLineHardware', 'OpeningPortException', 'RawFrame', 'ReadingException', 'SendingException', 'TimeoutException']



import time
from sys import platform

from scapy.config import conf

if platform.startswith('win32'):
	conf.contribs['ISOTP'] = {'use-can-isotp-kernel-module': False}
else:
	conf.contribs['ISOTP'] = {'use-can-isotp-kernel-module': True}
from scapy.contrib.isotp import ISOTP, ISOTPSocket

from ..hardware.can_hardware import CanFilter
from ..hardware.hardware_abc import HardwareABC, TimeoutException
from .transport_abc import PacketDirection, RawPacket, TransportABC


class Kwp2000OverCanTransport (TransportABC):
	def __init__ (self, hardware: HardwareABC, tx_id: int, rx_id: int) -> None:
		self.hardware = hardware
		self.tx_id, self.rx_id = tx_id, rx_id
		self.isotp: ISOTPSocket | None = None

	def init (self) -> bool:
		self.hardware.set_filters([CanFilter(can_id=self.rx_id, can_mask=0x7ff)])
		
		if not self.hardware.is_open():
			self.hardware.open()

		if not self.isotp:
			self.isotp = ISOTPSocket(self.hardware.socket, self.tx_id, self.rx_id, padding=True)

		return True
		
	def send_pdu (self, pdu: bytes) -> int:
		data = pdu
		bytes_written = self.isotp.send(data)

		self.buffer_push(RawPacket(direction=PacketDirection.OUTGOING, data=data, timestamp=int(time.time()*1000)))

		return len(data) # isotp socket doesnt return how many bytes were written

	def read_pdu (self) -> bytes:
		try:
			frame = self.isotp.sniff(timeout=self.hardware.get_timeout(), count=1)[0]
		except IndexError:
			self.isotp.close() # close the background thread that sends flow control frames
			raise TimeoutException

		self.buffer_push(RawPacket(direction=PacketDirection.INCOMING, data=frame.data, timestamp=int(time.time() * 1000)))

		return frame.data

	def send_read_pdu (self, data: bytes) -> bytes:
		self.buffer_push(RawPacket(direction=PacketDirection.OUTGOING, data=data, timestamp=int(time.time()*1000)))
		response = self.isotp.sr1(ISOTP(bytes(data)), verbose=False, timeout=self.hardware.get_timeout())

		if not response:
			raise TimeoutException

		self.buffer_push(RawPacket(direction=PacketDirection.INCOMING, data=response.data, timestamp=int(time.time() * 1000)))

		return response.data


from ctypes import LittleEndianStructure


class LittleEndianBitfield(LittleEndianStructure):
	_pack_ = 1

	def _to_dict (self) -> dict:
		return dict((field, getattr(self, field)) for field, _, _ in self._fields_ if getattr(self, field) != 0)

	def __str__ (self) -> str:
		return '{}({})'.format(
			type(self).__name__,
			', '.join([f'{k}={v}' for k,v in self._to_dict().items()])
		)

def ms_to_ns (miliseconds: int) -> int:
	return miliseconds*1000000

def ns_to_ms (nanoseconds: int) -> int:
	return round(nanoseconds/1000000)


from enum import Enum


class Kwp2000NegativeStatusIdentifierEnum(Enum):
	GENERAL_REJECT = 0x10
	SERVICE_NOT_SUPPORTED = 0x11
	SUB_FUNCTION_NOT_SUPPORTED_INVALID_FORMAT = 0x12
	INCORRECT_MESSAGE_LENGTH_OR_INVALID_FORMAT = 0x13
	RESPONSE_TOO_LONG = 0x14
	BUSY_REPEAT_REQUEST = 0x21
	CONDITIONS_NOT_CORRECT_OR_REQUEST_SEQUENCE_ERROR = 0x22
	ROUTINE_NOT_COMPLETE = 0x23
	REQUEST_SEQUENCE_ERROR = 0x24
	REQUEST_OUT_OF_RANGE = 0x31
	SECURITY_ACCESS_DENIED_SECURITY_ACCESS_REQUESTED = 0x33
	INVALID_KEY = 0x35
	EXCEED_NUMBER_OF_ATTEMPTS = 0x36
	REQUIRED_TIME_DELAY_NOT_EXPIRED = 0x37
	DOWNLOAD_NOT_ACCEPTED = 0x40
	IMPROPER_DOWNLOAD_TYPE = 0x41
	CANT_DOWNLOAD_TO_SPECIFIC_ADDRESS = 0x42
	CANT_DOWNLOAD_REQUESTED_NUMBER_OF_BYTES = 0x43
	UPLOAD_NOT_ACCEPTED = 0x50
	IMPROPER_UPLOAD_TYPE = 0x51
	CANT_UPLOAD_FROM_SPECIFIED_ADDRESS = 0x52
	CANT_UPLOAD_REQUESTED_NUMBER_OF_BYTES = 0x53
	UPLOAD_DOWNLOAD_NOT_ACCEPTED = 0x70
	TRANSFER_SUSPENDED = 0x71
	GENERAL_PROGRAMMING_FAILURE = 0x72
	WRONG_BLOCK_SEQUENCE_COUNTER = 0x73
	ILLEGAL_ADDRESS_IN_BLOCK_TRANSFER = 0x74
	ILLEGAL_BYTE_COUNT_IN_BLOCK_TRANSFER = 0x75
	ILLEGAL_BLOCK_TRANSFER_TYPE = 0x76
	BLOCK_TRANSFER_DATA_CHECKSUM_ERROR = 0x77
	REQUEST_CORRECTLY_RECEIVED_RESPONSE_PENDING = 0x78
	INCORRECT_BYTE_COUNT_DURING_BLOCK_TRANSFER = 0x79
	SUBFUNCTION_NOT_SUPPORTED_IN_ACTIVE_SESSION = 0x7E
	SERVICE_NOT_SUPPORTED_IN_CURRENT_SESSION = 0x7F
	SERVICE_NOT_SUPPORTED_IN_ACTIVE_DIAGNOSTIC_SESSION = 0x80
	RPM_TOO_HIGH = 0x81
	RPM_TOO_LOW = 0x82
	ENGINE_IS_RUNNING = 0x83
	ENGINE_IS_NOT_RUNNING = 0x84
	ENGINE_RUN_TIME_TOO_LOW = 0x85
	TEMPERATURE_TOO_HIGH = 0x86
	TEMPERATURE_TOO_LOW = 0x87
	VEHICLE_SPEED_TOO_HIGH = 0x88
	VEHICLE_SPEED_TOO_LOW = 0x89
	THROTTLE_PEDAL_TOO_HIGH = 0x8A
	THROTTLE_PEDAL_TOO_LOW = 0x8B
	TRANSMISSION_RANGE_NOT_IN_NEUTRAL = 0x8C
	TRANSMISSION_RANGE_NOT_IN_GEAR = 0x8D
	BRAKE_SWITCH_ES_NOT_CLOSED_PEDAL_NOT_APPLIED = 0x8F
	SHIFTER_LEVER_NOT_IN_PARK = 0x90
	TORQUE_CONVERTER_CLUTCH_LOCKED = 0x91
	VOLTAGE_TOO_HIGH = 0x92
	VOLTAGE_TOO_LOW = 0x93
	DATA_DECOMPRESSION_FAILED = 0x9A
	DATA_DECRYPTION_FAILED = 0x9B
	ECU_NOT_RESPONDING = 0xA0
	ECU_ADDRESS_UNKNOWN = 0xA1

class Kwp2000NegativeStatusMessageEnum(Enum):
	GENERAL_REJECT = 'General Reject'
	SERVICE_NOT_SUPPORTED = 'Service Not Supported'
	SUB_FUNCTION_NOT_SUPPORTED_INVALID_FORMAT = 'Sub Function Not Supported / Invalid Format'
	INCORRECT_MESSAGE_LENGTH_OR_INVALID_FORMAT = 'Incorrect message length or invalid format'
	RESPONSE_TOO_LONG = 'Response too long'
	BUSY_REPEAT_REQUEST = 'Busy / Repeat Request'
	CONDITIONS_NOT_CORRECT_OR_REQUEST_SEQUENCE_ERROR = 'Conditions Not Correct Or Request Sequence Error'
	ROUTINE_NOT_COMPLETE = 'Routine Not Complete'
	REQUEST_SEQUENCE_ERROR = 'Request sequence error'
	REQUEST_OUT_OF_RANGE = 'Request out of range'
	SECURITY_ACCESS_DENIED_SECURITY_ACCESS_REQUESTED = 'Security Access Denied / Security Access Requested'
	INVALID_KEY = 'Invalid Key'
	EXCEED_NUMBER_OF_ATTEMPTS = 'Exceed number of attempts'
	REQUIRED_TIME_DELAY_NOT_EXPIRED = 'Required time delay not expired'
	DOWNLOAD_NOT_ACCEPTED = 'Download not accepted'
	IMPROPER_DOWNLOAD_TYPE = 'Improper download type'
	CANT_DOWNLOAD_TO_SPECIFIC_ADDRESS = "Can't download to specific address"
	CANT_DOWNLOAD_REQUESTED_NUMBER_OF_BYTES = "Can't download requested number of bytes"
	UPLOAD_NOT_ACCEPTED = 'Upload not accepted'
	IMPROPER_UPLOAD_TYPE = 'Improper upload type'
	CANT_UPLOAD_FROM_SPECIFIED_ADDRESS = "Can't upload from specified address"
	CANT_UPLOAD_REQUESTED_NUMBER_OF_BYTES = "Can't upload requested number of bytes"
	UPLOAD_DOWNLOAD_NOT_ACCEPTED = 'Upload/download not accepted'
	TRANSFER_SUSPENDED = 'Transfer suspended'
	GENERAL_PROGRAMMING_FAILURE = 'General programming failure'
	WRONG_BLOCK_SEQUENCE_COUNTER = 'Wrong block sequence counter'
	ILLEGAL_ADDRESS_IN_BLOCK_TRANSFER = 'Illegal address in block transfer'
	ILLEGAL_BYTE_COUNT_IN_BLOCK_TRANSFER = 'Illegal byte count in block transfer'
	ILLEGAL_BLOCK_TRANSFER_TYPE = 'Illegal block transfer type'
	BLOCK_TRANSFER_DATA_CHECKSUM_ERROR = 'Block transfer data checksum error'
	REQUEST_CORRECTLY_RECEIVED_RESPONSE_PENDING = 'Request correctly received / Response pending'
	INCORRECT_BYTE_COUNT_DURING_BLOCK_TRANSFER = 'Incorrect byte count during block transfer'
	SUBFUNCTION_NOT_SUPPORTED_IN_ACTIVE_SESSION = 'Subfunction not supported in active session'
	SERVICE_NOT_SUPPORTED_IN_CURRENT_SESSION = 'Service not supported in current session'
	SERVICE_NOT_SUPPORTED_IN_ACTIVE_DIAGNOSTIC_SESSION = 'Service not supported in active diagnostic session'
	RPM_TOO_HIGH = 'RPM too high'
	RPM_TOO_LOW = 'RPM too low'
	ENGINE_IS_RUNNING = 'Engine is running'
	ENGINE_IS_NOT_RUNNING = 'Engine is not running'
	ENGINE_RUN_TIME_TOO_LOW = 'Engine run time too low'
	TEMPERATURE_TOO_HIGH = 'Temperature too high'
	TEMPERATURE_TOO_LOW = 'Temperature too low'
	VEHICLE_SPEED_TOO_HIGH = 'Vehicle speed too high'
	VEHICLE_SPEED_TOO_LOW = 'Vehicle speed too low'
	THROTTLE_PEDAL_TOO_HIGH = 'Throttle/pedal too high'
	THROTTLE_PEDAL_TOO_LOW = 'Throttle/pedal too low'
	TRANSMISSION_RANGE_NOT_IN_NEUTRAL = 'Transmission range not in neutral'
	TRANSMISSION_RANGE_NOT_IN_GEAR = 'Transmission range not in gear'
	BRAKE_SWITCH_ES_NOT_CLOSED_PEDAL_NOT_APPLIED = 'Brake switch(es) not closed (pedal not applied)'
	SHIFTER_LEVER_NOT_IN_PARK = 'Shifter lever not in park'
	TORQUE_CONVERTER_CLUTCH_LOCKED = 'Torque converter clutch locked'
	VOLTAGE_TOO_HIGH = 'Voltage too high'
	VOLTAGE_TOO_LOW = 'Voltage too low'
	DATA_DECOMPRESSION_FAILED = 'Data decompression failed'
	DATA_DECRYPTION_FAILED = 'Data decryption failed'
	ECU_NOT_RESPONDING = 'ECU not responding'
	ECU_ADDRESS_UNKNOWN = 'ECU address unknown'

class Kwp2000NegativeStatus:
	identifier: int = 0x0
	name: str = '' 
	message: str = ''

	def __init__ (self, identifier: int) -> None:
		self.identifier = identifier
		try:
			self.name = Kwp2000NegativeStatusIdentifierEnum(self.identifier).name
		except ValueError:
			return

		try:
			self.message = Kwp2000NegativeStatusMessageEnum[self.name].value
		except KeyError:
			self.message = 'Unknown'

	def get_identifier (self) -> int:
		return self.identifier

	def get_name (self) -> str:
		return self.name

	def get_message (self) -> str:
		return self.message

	def __str__ (self) -> str:
		return 'Kwp2000NegativeStatus(identifier={},message={})'.format(hex(self.identifier), self.message)


from ctypes import c_uint8

from ...utils import LittleEndianBitfield


class ResourceMaskBitfield(LittleEndianBitfield):
	_fields_ = [ # res - reserved
		('CAL', c_uint8, 1),
		('DAQ', c_uint8, 1),
		('res', c_uint8, 1),
		('res1', c_uint8, 1),
		('res2', c_uint8, 1),
		('res3', c_uint8, 1),
		('PGM', c_uint8, 1),
		('res4', c_uint8, 1)
	]

class SessionStatusBitfield(LittleEndianBitfield):
	_fields_ = [ # res - reserved
		('CAL', c_uint8, 1),
		('DAQ', c_uint8, 1),
		('RESUME', c_uint8, 1),
		('res', c_uint8, 1),
		('res1', c_uint8, 1),
		('res2', c_uint8, 1),
		('STORE', c_uint8, 1),
		('RUN', c_uint8, 1),
	]


import time

from ..hardware.can_hardware import CanFilter
from ..hardware.hardware_abc import HardwareABC, RawFrame
from .transport_abc import PacketDirection, RawPacket, TransportABC


class CcpOverCanTransport (TransportABC):
	'''
	CAN Calibration Protocol transport. 
	Nothing interesting here - CCP doesn't use ISOTP or any other way to break 
	chunks of data into messages - each command or response must fit within 8 bytes of
	a standard CAN frame
	'''
	def __init__ (self, hardware: HardwareABC, tx_id: int, rx_id: int) -> None:
		self.hardware = hardware
		self.tx_id, self.rx_id = tx_id, rx_id
		self.hardware.set_filters([CanFilter(can_id=self.rx_id, can_mask=0x7ff)])
		
	def send_pdu (self, pdu: bytes) -> int:
		data = pdu
		bytes_written = self.hardware.write(RawFrame(identifier=self.tx_id, data=data))

		self.buffer_push(RawPacket(direction=PacketDirection.OUTGOING, data=data, timestamp=int(time.time()*1000)))

		return len(data) # can socket doesnt return how many bytes were written @TODO: verify

	def read_pdu (self) -> bytes:
		frame = self.hardware.read(8).data

		self.buffer_push(RawPacket(direction=PacketDirection.INCOMING, data=frame, timestamp=int(time.time() * 1000)))

		return frame


from abc import ABC

from typing_extensions import Self

from ..transport.transport_abc import TransportABC


class ProtocolException(IOError):
	pass

class ProtocolABC (ABC):
	def __init__ (self, transport: TransportABC) -> None:
		self.transport = transport

	def open (self) -> bool:
		'''
		Initialize hardware/transport layers - get ready to interact
		'''
		pass

	def init (self) -> bool:
		'''
		Initialize the protocol layer
		'''

	def execute (self, command):
		'''
		Execute a command and return the results
		'''
		pass

	def close (self) -> None:
		'''
		Shutdown the underlaying layers
		'''
		pass

	def set_transport (self, transport: TransportABC) -> Self:
		'''
		Replace transport used by the protocol instance. 
		Remember to close the previous socket first, 
		this won't happen automatically

		:param transport: Transport instance to replace current one with
		'''
		self.transport = transport

	def __del__ (self) -> None:
		self.close()


from enum import Enum


class TimingParameterIdentifier(Enum):
	READ_LIMITS_OF_POSSIBLE_TIMING_PARAMETERS = 0x00
	SET_TIMING_PARAMETERS_TO_DEFAULT_VALUES = 0x01
	READ_CURRENTLY_ACTIVE_TIMING_PARAMETERS = 0x02
	SET_TIMING_PARAMETERS_TO_GIVEN_VALUES = 0x03

class DiagnosticSession(Enum):
	DEFAULT = 0x81
	FLASH_REPROGRAMMING = 0x85
	ENGINEERING = 0x86
	ADJUSTMENT = 0x87
	STANDBY = 0x89
	PASSIVE = 0x90
	EXTENDED_DIAGNOSTIC = 0x92

class AccessType(Enum):
	PROGRAMMING_REQUEST_SEED = 0x01
	PROGRAMMING_SEND_KEY = 0x02

class ResetMode(Enum):
	POWER_ON_RESET = 0x01
	NONVOLATILE_MEMORY_RESET = 0x82

class InputOutputControlParameter(Enum):
	RETURN_CONTROL_TO_ECU = 0x00
	REPORT_CURRENT_STATE = 0x01
	RESET_TO_DEFAULT = 0x04
	FREEZE_CURRENT_STATE = 0x05
	SHORT_TERM_ADJUSTMENT = 0x07
	LONG_TERM_ADJUSTMENT = 0x08

class CompressionType(Enum):
	UNCOMPRESSED = 0x0

class EncryptionType(Enum):
	UNENCRYPTED = 0x0

class ResponseType(Enum):
	REQUIRED = 0x01
	NOT_REQUIRED = 0x02

class DtcStatus(Enum):
	REQUEST_IDENTIFIED_DTC_AND_STATUS = 0x00
	REQUEST_SUPPORTED_DTC_AND_STATUS = 0x01
	REQUEST_IDENTIFIED_TWO_BYTE_HEX_DTC_AND_STATUS = 0x02
	REQUEST_SUPPORTED_TWO_BYTE_HEX_DTC_AND_STATUS = 0x03
	REQUEST_MOST_RECENT_DTC = 0x04
	REQUEST_EXTENDED_NUMBER_OF_SUPPORTED_DTCS = 0xE0

class DtcGroup(Enum):
	POWERTRAIN = 0x0000
	CHASSIS = 0x4000
	BODY = 0x8000
	NETWORK = 0xC000
	ALL = 0xFF00


from typing import Union

from .ccp_command import CcpCommand
from .enums import DataTransmissionMode, DataTransmissionRequest, DisconnectType
from .types import ResourceMaskBitfield, SessionStatusBitfield


class Connect(CcpCommand):
	'''
	Estabilish a master-slave connection
	
	:param station_address: station address, word. MSB
	'''
	code = 0x01

	def init (self, station_address: int) -> None:
		self.data = station_address.to_bytes(2, 'little')

class ExchangeStationIdentifications(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	The CCP master and slave stations exchange IDs for automatic session
	configuration. This might include automatic assignment of a data acquisition
	setup file depending on the slave's returned ID (Plug´n´Play). 
	The slave device automatically sets the Memory Transfer Address 0 (MTA0) to
	the location from which the CCP master may subsequently upload the requested
	ID using UPLOAD. See also the SET_MTA and UPLOAD command
	description.
	'''
	code = 0x17

class GetSeedForKey(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	See EXCHANGE_ID for a description of the resource mask.
	Only one resource or function may be requested with one GET_SEED
	command If more than one resource is requested, the GET_SEED command
	together with the following UNLOCK command has to be performed multiple
	times.
	Returns ´seed´ data for a seed&key algorithm for computing the ´key´ to
	unlock the requested function for authorized access (see 'Unlock Protection'
	below).

	:param resource_mask: Requested slave resource or function
	'''
	code = 0x12

	def init (self, resource_mask: Union[int, ResourceMaskBitfield]) -> None:
		if isinstance(resource_mask, ResourceMaskBitfield):
			self.data = bytes(resource_mask)
		else:
			self.data = resource_mask.to_bytes(1, 'little')
		
class UnlockProtection(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	Unlocks the slave devices security protection (if applicable) using a ´key´ computed from
	´seed´. See seed&key above

	:param key: key calculated from the seed. Up to 6 bytes
	'''
	code = 0x13

	def init (self, key: bytes) -> None:
		self.data = key

class SetMemoryTransferAddress(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
		
	This command will initialize a base pointer (32Bit + extension) for all following memory
	transfers. The address extension depends on the slave controller's organization and may
	identify a switchable memory bank or a memory segment.
	The MTA number (handle) is used to identify different transfer address locations (pointers).
	MTA0 is used by the commands DNLOAD, UPLOAD, DNLOAD_6 , SELECT_CAL_PAGE,
	CLEAR_MEMORY, PROGRAM and PROGRAM_6. MTA1 is used by the MOVE command.
	See also command ‘MOVE’.

	:param mta_number: 0 or 1. 
		Used to identify pointer locations; 
		0 for download, upload, select_cal_page, clear_memory, program
		1 for move
	:param address_extension: 1 byte address extension
	:param address: unsigned long address
	'''
	code = 0x02

	def init (self, mta_number: int, address_extension: int, address: int) -> None:
		if mta_number not in [0, 1]:
			raise ValueError(f'Invalid MTA number: {mta_number}. Valid values are either 0 or 1')

		self.data = bytes([mta_number, address_extension]) + address.to_bytes(4, 'little')

class DataDownload(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	The data block of the specified length (size) contained in the CRO will be copied into
	memory, starting at the current Memory Transfer Address 0 (MTA0). The MTA0 pointer will
	be post-incremented by the value of 'size'.

	:param size: size of the data block about to be transfered (max 5 bytes)
	:param data: data to be transferred
	'''
	code = 0x03

	def init (self, size: int, data: bytes) -> None:
		if (size > 5):
			raise ValueError(f'Invalid size: {size}. Max value is 5')

		if (len(data) > 5):
			raise ValueError('Invalid data size. Max length is 5')

		self.data = size.to_bytes(1, 'little') + data

class DataDownload6Bytes(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	The data block with the fixed length (size) of 6 bytes contained in the CRO will be copied
	into memory, starting at the current Memory Transfer Address 0 (MTA0). The MTA0 pointer
	will be post-incremented by the value 6.

	:param data: 6 bytes of data to be transferred
	'''
	code = 0x23

	def init (self, data: bytes) -> None:
		if (len(data) > 6):
			raise ValueError('Invalid data size. Max length is 6')

		self.data = data

class DataUpload(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	A data block of the specified length (size), starting at current MTA0, will be copied into the
	corresponding DTO data field. The MTA0 pointer will be post-incremented by the value of
	'size'.

	:param size: size of data block to be uploaded - maximum 5
	'''
	code = 0x04

	def init (self, size: int) -> None:
		if (size > 5):
			raise ValueError(f'Invalid requested size: {size}. Maximum value is 5')

		self.data = size.to_bytes(1, 'little')

class ShortUpload(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	A data block of the specified length (size), starting at source address will be copied into the
	corresponding DTO data field. The MTA0 pointer remains unchanged.
	
	:param size: size of data block to be uploaded - maximum 5
	:param address_extension: byte address extension
	:param address: address - unsigned long
	'''
	code = 0x0F

	def init (self, size: int, address_extension: int, address: int) -> None:
		if (size > 5):
			raise ValueError(f'Invalid requested size: {size}. Maximum value is 5')

		self.data = bytes([size, address_extension]) + address.to_bytes(4, 'little')

class SelectCalibrationPage(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	This command's function depends on the ECU implementation. The previously initialized
	MTA0 points to the start of the calibration data page that is selected as the currently active
	page by this command.
	'''
	code = 0x11

class GetSizeOfDaqList(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	Returns the size of the specified DAQ list as the number of available Object
	DescriptorTables (ODTs) and clears the current list. If the specified list number is not
	available, size = 0 should be returned. The DAQ list is initialized and data acquisition by this
	list is stopped.
	An individual CAN Identifier may be assigned to a DAQ list to configure multi ECU data
	acquisition. This feature is optional. If the given identifier isn’t possible, an error code is
	returned. 29 bit CAN identifiers are marked by the most significant bit set.

	:param list_number: DAQ list number (0, 1, ...)
	:param can_identifier: CAN identifier of DTO dedicated to list number. unsigned long
	'''
	code = 0x14

	def init (self, list_number: int, can_identifier: int = 0) -> None:
		self.data = list_number.to_bytes(1, 'little') + b'\xFF' + can_identifier.to_bytes(4, 'little')

class SetDaqListPointer(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	Initializes the DAQ list pointer for a subsequent write to a DAQ list.
	See also 'Organization of Data Acquisition Messages'.

	:param list_number: DAQ list number (0, 1, ...)
	:param odt_number: Object Descriptor Table ODT number (0, 1, ...)
	:param element_number: Element number within ODT (0, 1, ...)
	'''
	code = 0x15

	def init (self, list_number: int, odt_number: int, element_number: int) -> None:
		self.data = bytes([list_number, odt_number, element_number])

class WriteDaqListEntry(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	Writes one entry (description of single DAQ element) to a DAQ list defined by the DAQ list
	pointer (see SET_DAQ_PTR). The following DAQ element sizes are defined: 1 byte , 2
	bytes (type word), 4 bytes (type long / Float).
	An ECU may not support individual address extensions for each element and 2 or 4 byte
	element sizes. It is the responsibility of the master device to care for the ECU limitations.
	The limitations may be defined in the slave device description file (e.g. ASAP2).
	It is the responsibility of the slave device, that all bytes of a DAQ element are consistent
	upon transmission.
	See also 'Organization of Data Acquisition Messages'.

	:param size: Size of DAQ element in bytes (1, 2, 4)
	:param address_extension: Address extension of DAQ element
	:param address: Address of DAQ element - unsigned long
	'''
	code = 0x16

	def init (self, size: int, address_extension: int, address: int) -> None:
		self.data = bytes([size, address_extension]) + address.to_bytes(4, 'little')

class StartStopDataTransmission(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	This command is used to start or to stop the data acquisition or to prepare a synchronized
	start of the specified DAQ list. The Last ODT number specifies which ODTs (From 0 to
	Last ODT number) of this DAQ list should be transmitted. The Event Channel No.
	specifies the generic signal source that effectively determines the data transmission timing.
	To allow reduction of the desired transmission rate, a prescaler may be applied to the
	Event Channel. The prescaler value factor must be greater than or equal to 1.
	The mode parameter is defined as follows: 0x00 stops specified DAQ list, 0x01 starts
	specified DAQ list, 0x02 prepares DAQ list for synchronised start.
	The start/stop mode parameter = 0x02 (prepare to start data transmission) configures the
	DAQ list with the provided parameters but does not start the data acquisition of the
	specified list. This parameter is used for a synchronized start of all configured DAQ lists. In
	case the slave device is not capable of performing the synchronized start of the data
	acquisition, the slave device may then start data transmission if this parameter is TRUE
	(not zero).
	The ECU specific properties of event channels and DAQ lists may be described in the slave
	device description (ASAP2).

	:param mode: start / stop / prepare data transmission
	:param daq_list_number: DAQ list number
	:param last_odt_number: Last ODT number
	:param event_channel: Event channel number
	:param prescaler: Transmission rate prescaler - word
	'''
	code = 0x06

	def init (self, mode: DataTransmissionMode, daq_list_number: int, last_odt_number: int, event_channel: int, prescaler: int) -> None:
		self.data = bytes([mode.value, daq_list_number, last_odt_number, event_channel]) + prescaler.to_bytes(2, 'little')

class Disconnect(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	Disconnects the slave device. The disconnect can be temporary, setting the slave device in
	an "off line" state or with parameter 0x01 terminating the calibration session.
	Terminating the session invalidates all state information and resets the slave protection
	status.
	A temporary disconnect doesn’t stop the transmission of DAQ messages. The MTA values,
	the DAQ setup, the session status and the protection status are unaffected by the
	temporary disconnect and remain unchanged.
	If the ECU supports the resume feature and the resume bit was set with a
	SET_SESSION_STATUS command, the DAQ related functions behave like in a temporary
	disconnect. The protections status for DAQ remains unlocked.

	:param disconnect_type: Temporary or end of session
	:param station_address: station address, word
	'''
	code = 0x07

	def init (self, disconnect_type: DisconnectType, station_address: int) -> None:
		self.data = disconnect_type.value.to_bytes(1, 'little') + station_address.to_bytes(2, 'little')

class SetSessionStatus(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	Keeps the slave node informed about the current state of the calibration session (see also
	chapter 'Error Handling').

	Bits are set (1) if expression is TRUE.
	The session status bits are read/write to the slave device and are be cleared on power-up,
	on session log-off and in applicable fault conditions.

	Session status, LSB:
	RUN | STORE | res | res | res | RESUME | DAQ | CAL

	bit 0: CAL - Calibration data initialized
	bit 1: DAQ - DAQ lists initialized 
	bit 2: RESUME - Request to save DAQ setup during shutdown in ECU. ECU automatically restarts DAQ after startup
	bit 3: res - reserved
	bit 4: res - reserved
	bit 5: res - reserved
	bit 6: STORE - Request to save calibration data during shut-down in ECU
	bit 7: RUN - Session in progress 

	:param status_bits: Session status bits. See table above 
	'''
	code = 0x0C

	def init (self, status: Union[int, SessionStatusBitfield]) -> None:
		if isinstance(status, SessionStatusBitfield):
			status = bytes(status)
		else:
			status = status.to_bytes(1, 'little')

		self.data = status

class GetSessionStatus(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	Note: the use of additional status information is manufacturer and / or project specific, it
	is not part of this protocol specification. For example, additional status information could
	contain an incremental checksum result, that keeps track of the current session activities.
	If the return information does not contain additional status information, the additional
	status information qualifier has to be FALSE (0). If the additional status information is not
	FALSE, it may be used to determine the type of additional status information
	'''
	code = 0x0D

class BuildChecksum(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	Returns a checksum result of the memory block that is defined by MTA0
	(memory transfer area start address) and block size. The checksum algorithm
	may be manufacturer and / or project specific, it is not part of this specification.

	:param size: block size in bytes - unsigned long
	'''
	code = 0x0E

	def init (self, size: int) -> None:
		self.data = size.to_bytes(4, 'little') # @todo not sure about LSB here

class ClearMemory(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	This command may be used to erase FLASH EPROM prior to reprogramming. The MTA0
	pointer points to the memory location to be erased.

	:param size: memory size - long
	'''
	code = 0x10

	def init (self, size: int) -> None:
		self.data = size.to_bytes(4, 'little')

class Program(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	The data block of the specified length (size) contained in the CRO will be programmed into
	non-volatile memory (FLASH, EEPROM), starting at current MTA0. The MTA0 pointer will
	be post-incremented by the value of ‘size’.

	:param size: size of data block to follow (max 5) 
	:param data: data to be programmed
	'''
	code = 0x18

	def init (self, size: int, data: bytes) -> None:
		if size > 5:
			raise ValueError(f'Invalid requested size: {size}. Maximum value is 5')

		if len(data) > 5:
			raise ValueError('Invalid data size. Maximum length is 5')

		self.data = size.to_bytes(1, 'little') + data

class Program6Bytes(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	The data block with the length (size) of 6 bytes contained in the CRO will be programmed
	into non-volatile memory (FLASH, EEPROM), starting at current MTA0. The MTA0 pointer
	will be post-incremented by 6.

	:param data: data to be programmed (6 bytes)
	'''
	code = 0x22

	def init (self, data: bytes) -> None:
		if len(data) > 6:
			raise ValueError('Invalid data size. Maximum length is 6')

		self.data = data

class MoveMemoryBlock(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	A data block of the specified length (size) will becopied from the address defined by MTA 0
	(source pointer) to the address defined by MTA 1 (destination pointer).

	:param size: number of bytes to be moved (size of the data block) - long
	'''
	code = 0x19

	def init (self, size: int) -> None:
		self.data = size.to_bytes(4, 'little')

class DiagnosticService(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	The slave device carries out the requested service and automatically sets the Memory
	Transfer Address MTA0 to the location from which the CCP master device (host) may
	subsequently upload the requested diagnostic service return information.

	:param service_number: Diagnostic service number - word
	'''
	code = 0x20

	def init (self, service_number: int) -> None:
		self.data = service_number.to_bytes(2, 'little')

class ActionService(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	The slave device carries out the requested service and automatically sets the Memory
	Transfer Address MTA0 to the location from which the CCP master device may
	subsequently upload the requested action service return information (if applicable).

	:param service_number: Action service number - word
	:param parameters: parameters, if applicable
	'''
	code = 0x21 

	def init (self, service_number: int, parameters: bytes = bytes()) -> None:
		self.data = service_number.to_bytes(2, 'little') + parameters

class TestAvailability(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	This command is used to test if the slave with the specified station address is
	available for CCP communication. This command does not establish a logical
	connection nor does it trigger any activities in the specified slave. Station
	address is specified as a number in little-endian byte order (Intel format, low
	byte first).

	:param station_address: station address, word. 
	'''
	code = 0x05

	def init (self, station_address: int) -> None:
		self.data = station_address.to_bytes(2, 'little')

class StartStopSynchronisedDataTransmission(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:
	
	This command is used to start the periodic transmission of all DAQ lists
	configured with the previously sent START_STOP command (start/stop modus
	= 2) as „prepared to start“ in a synchronized manner. The command is used to
	stop the periodic transmission of all DAQ lists, including those not started
	synchronized.

	:param request: Stop/Start data transmission
	'''
	code = 0x08

	def init (self, request: DataTransmissionRequest) -> None:
		self.data = request.value.to_bytes(1, 'little')

class GetCurrentlyActiveCalibrationPage(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	This command returns the start address of the calibration page that is currently active in the
	slave device.
	'''
	code = 0x09

class GetImplementedVersionOfCcp(CcpCommand):
	'''
	Copied placeholder, to be translated from ISO to human:

	This command performs a mutual identification of the protocol version used in the
	master and in the slave device to aggree on a common protocol version. This
	command is expected to be executed prior to the EXCHANGE_ID command.

	:param main_protocol_version: Desired main protocol version
	:param release_protocol_version: Desired release within version
	'''
	code = 0x1B

	def init (self, main_protocol_version: int, release_protocol_version: int) -> None:
		self.data = bytes([main_protocol_version, release_protocol_version])


import logging
from dataclasses import dataclass

from ...transport import TransportABC
from ..protocol_abc import ProtocolABC, ProtocolException
from .ccp_command import CcpCommand
from .ccp_response import CcpResponse, CcpResponseFrame, CcpReturnCode

logger = logging.getLogger(__name__)

class CcpException(ProtocolException):
	pass

class CcpNegativeResponseException(CcpException):
	def __init__ (self, return_code: CcpReturnCode) -> None:
		self.return_code = return_code

	def __str__ (self) -> str:
		return '{}: {} - {}'.format(self.return_code.category, hex(self.return_code.value), self.return_code.name)
	
	def __repr__ (self) -> str:
		return self.__str__()

@dataclass
class CcpRequestFrame:
	'''
	Also known as Command Receive Object - CRO. 
	Perhaps this should be renamed

	Structure: 
	1 byte - command code 
	1 byte - counter of transmitted frames
	6 bytes - command data
	'''
	command_code: int
	counter: int
	data: bytes

	def to_pdu (self) -> bytes:
		data = self.data + bytes([0x00]*(6-len(self.data))) # padding
		return bytes([self.command_code, self.counter]) + data

class CcpProtocol (ProtocolABC):
	'''
	CAN Calibration Protocol
	'''
	def __init__ (self, transport: TransportABC, byte_order: str = 'little'):
		'''
		CCP protocol doesn't specify byte order for parameters - it depends on the receiving ECU

		:param byte_order: 'little' or 'big'
		'''
		self.transport = transport
		self.byte_order = byte_order
		self.command_counter = 0
		if byte_order != 'little':
			raise NotImplementedError('Byte order switching isn\'t implemented yet. You can resort to manually setting data attribute on commands')

	def open (self) -> bool:
		return self.transport.hardware.open()

	def execute (self, command: CcpCommand) -> CcpResponse:
		frame = CcpRequestFrame(command.get_code(), self.command_counter, command.get_data())

		response_pdu = self.transport.send_read_pdu(data=frame.to_pdu())
		self.command_counter += 1

		if (self.command_counter > 0xFF):
			self.command_counter = 0x00

		response = CcpResponse(
			CcpReturnCode(response_pdu[1]),
			CcpResponseFrame(
				packet_id=response_pdu[0], 
				status=response_pdu[1],
				counter=response_pdu[2],
				data=response_pdu[3:]
				)
			)

		response = self.handle_errors(response)

		return response

	def handle_errors (self, response: CcpResponse) -> CcpResponse:
		if response.success():
			return response

		raise CcpNegativeResponseException(response.return_code)

	def close (self) -> None:
		self.transport.hardware.close()


import logging
import time

from ..hardware.hardware_abc import HardwareABC, RawFrame
from .transport_abc import PacketDirection, RawPacket, TransportABC

logger = logging.getLogger(__name__)

class Kwp2000OverKLineTransport (TransportABC):
	def __init__ (self, hardware: HardwareABC, tx_id: int, rx_id: int) -> None:
		self.hardware: HardwareABC = hardware
		self.tx_id, self.rx_id = tx_id, rx_id

	def send_pdu (self, pdu: bytes) -> int:
		data = self.build_payload(pdu)

		bytes_written = self._write(data)
		self.buffer_push(RawPacket(direction=PacketDirection.OUTGOING, data=data, timestamp=int(time.time()*1000)))

		return bytes_written

	def read_pdu (self) -> bytes:
		frame = bytes()
		data = bytes()

		frame += self._read(4)

		counter = frame[0]
		tx_rx_id = frame[1:3]

		if (counter == 0x80): # more than 127 bytes incoming, counter overflowed. counter is gonna come after IDs
			counter = frame[3]+1
			logger.debug('More than 127 bytes incoming: {}'.format(counter))
		else:
			counter = counter-0x80
			data += frame[3:4] # frame[3] == python would automatically convert it to int. frame[3:3] == byte slice 

		data_and_checksum = self._read(counter)
		frame += data_and_checksum
		data += data_and_checksum[:-1] # last byte is the checksum

		self.buffer_push(RawPacket(direction=PacketDirection.INCOMING, data=frame, timestamp=int(time.time() * 1000)))

		return data

	def _write (self, data: bytes) -> int:
		logger.debug('K-Line sending: {}'.format(' '.join([hex(x) for x in list(data)])))
		return self.hardware.write(RawFrame(identifier=0, data=data))

	def _read (self, length: int) -> bytes:
		logger.debug('K-Line trying to read {} bytes'.format(length))
		data = self.hardware.read(length).data
		logger.debug('K-Line success: {}'.format(' '.join([hex(x) for x in list(data)])))
		return data

	def init (self, payload: bytes) -> list[tuple[bytes, int, int]]:
		'''
		Bring up the socket if not opened already and initialize the K-Line by ISO14230. 
		'''
		if not self.hardware.is_open():
			self.hardware.open()

		init_payload = self.build_payload(payload)
		responses = []
		responses.append(self.hardware.iso14230_fast_init(init_payload, timing_offset_ms=0))
		# @todo: find out if there is some way to self-test and automatically measure the timing offset we should apply
		if responses[0][0][0:1] not in [b'\x00', b'\x81', b'\xC1']: # most often seen responses, depends on the adapter
			responses.append(self.hardware.iso14230_fast_init(init_payload, timing_offset_ms=-2))
			responses.append(self.hardware.iso14230_fast_init(init_payload, timing_offset_ms=2))

		return responses

	def calculate_checksum (self, payload: bytes) -> int:
		return sum(payload) & 0xFF

	def build_payload (self, data: bytes) -> bytes:
		data_length = len(data)

		if (data_length < 127):
			counter = 0x80 + data_length
		else:
			counter = 0x80
			data = bytes([data_length]) + data

		payload = bytes([counter, self.tx_id, self.rx_id]) 
		payload += data
		payload += self.calculate_checksum(payload).to_bytes(1, 'big')

		return payload


from enum import Enum


class DisconnectType(Enum):
	TEMPORARY = 0x00 
	END_OF_SESSION = 0x01

class DataTransmissionRequest(Enum):
	STOP = 0x00
	START = 0x01

class DataTransmissionMode(Enum):
	START = 0x00
	STOP = 0x01
	PREPARE = 0x02


import os
import time
from dataclasses import dataclass
from sys import platform

from scapy.config import conf
from typing_extensions import Self

if platform.startswith('win32'):
	conf.contribs['CANSocket'] = {'use-python-can': True}
else:
	# If you're a linux guest in a windows host, this will cause
	# "no data received, try increasing stmin" issues, ranging 
	# from occasional to very common.
	# Best guess so far is that since python-can won't use the
	# native ISO-TP kernel module, the issue must be closely related
	# to some low level device access handling in the passthrough hypervisor
	#
	# Case in point used a Canable interface with Candlelight firmware,
	# Ubuntu 22.04 (5.15.0 kernel) guest on a Windows 10 v1803 host, 
	# through VMware Workstation 16
	conf.contribs['CANSocket'] = {'use-python-can': False}

	from scapy.contrib.cansocket import CANSocket
from scapy.error import Scapy_Exception
from scapy.layers.can import CAN, CAN_MAX_DLEN, CAN_MTU

from .hardware_abc import (
	HardwareABC,
	HardwarePort,
	OpeningPortException,
	RawFrame,
	TimeoutException,
)

CAN_HEADER_LEN = CAN_MTU-CAN_MAX_DLEN

@dataclass
class CanFilter:
	can_id: int
	can_mask: int = 0x7ff

class CanHardware(HardwareABC):
	'''
	Hardware class for CAN Bus interfaces. Uses Scapy as a backend

	:param filters: list of dictionaries containing hardware CAN filters.
		Keys: can_id, can_mask
	'''

	def __init__ (self, port: str, timeout: int = 1, filters: list[CanFilter] | None = None) -> None:
		self.port: str = port
		self.timeout: float = timeout
		self.filters: list[CanFilter] = filters if filters != None else []

	def _build_filters (self) -> list[dict]:
		filters = []
		
		for can_filter in self.filters:
			filters.append({'can_id': can_filter.can_id, 'can_mask': can_filter.can_mask})

		if len(filters) == 0:
			filters = None

		return filters

	def open (self) -> bool:
		try:
			self.socket = CANSocket(channel=self.port, can_filters=self._build_filters())
		except (OSError, Scapy_Exception) as e:
			raise OpeningPortException(e)
		return True

	def is_open (self) -> bool:
		try:
			return not self.socket.closed
		except (AttributeError, TypeError, ValueError):
			return False

	def read (self, length: int) -> RawFrame:
		try:
			packet = self.socket.sniff(timeout=self.timeout, count=1)[0]
		except IndexError:
			raise TimeoutException

		return RawFrame(identifier=packet.identifier, data=packet.data)

	def write (self, frame: RawFrame) -> int:
		packet = CAN(identifier=frame.identifier, data=frame.data)

		bytes_written = self.socket.send(packet)

		return bytes_written-CAN_HEADER_LEN

	def _read_with_software_filters (self, filters: list, timeout: float = 10.0) -> RawFrame:
		'''
		This should not be used and will be removed soon
		'''
		time_started = time.time()

		while True:
			packet = self.socket.recv()
			can_id = packet.identifier

			for can_filter in filters:
				if (can_id & can_filter['can_mask']) == (can_filter['can_id'] & can_filter['can_mask']):
					return RawFrame(identifier=packet.identifier, data=packet.data)

			if (time.time()-time_started) > timeout:
				raise TimeoutException

	def get_filters (self) -> list[CanFilter]:
		'''
		Get active hardware canbus id filters
		'''
		return self.filters

	def set_filters (self, filters: list[CanFilter]) -> Self:
		'''
		Replace hardware canbus id filters. 
		This will triger a restart of the socket
		'''
		self.filters = filters
		self.close()
		self.open()

	def add_filter (self, can_filter: CanFilter) -> Self:
		'''
		Add new hardware canbus id filter
		This will trigger a restart of the socket
		'''
		self.set_filters(self.get_filters() + can_filter)

	def close (self) -> None:
		if hasattr(self, 'socket'):
			self.socket.close()

	def set_baudrate (self, baudrate: int) -> Self:
		raise NotImplementedError

	def get_baudrate (self) -> int:
		raise NotImplementedError

	def set_timeout (self, timeout: float) -> Self:
		self.socket.timeout = timeout
		self.timeout = timeout
		return self

	@staticmethod
	def _available_ports_linux () -> list[str]:
		'''
		Returns a list of available CAN network interfaces by checking the /sys/class/net/<iface>/type file

		:return: a list containing interface identifiers (can0, vcan0, ...)
		'''
		NET_DIR = '/sys/class/net'
		interfaces: list[str] = []

		if not os.path.exists(NET_DIR):
			return interfaces

		for iface in os.listdir(NET_DIR):
			type_path = os.path.join(NET_DIR, iface, 'type')
			if os.path.isfile(type_path):
				with open(type_path, 'r') as f:
					iface_type = f.read().strip()

				if iface_type == '280':
					interfaces.append(iface)

		return interfaces

	@staticmethod
	def available_ports () -> list[HardwarePort]:
		if platform.startswith('win32'):
			return []
		else:
			ports = [HardwarePort(port=x, port_name=x) for x in CanHardware._available_ports_linux()]
		
		return ports
